"""
    This class holds methods used by:
        1. main.py in order to set up database at
        program start
        2. record_dao_local.py classes to read from and write to local database
"""


import sqlite3
import os
import os.path
import sys
from pkg_resources import resource_string

class LocalDBAccessor(object):
    def __init__(self, dbconfig):
        if not dbconfig:
            raise ValueError('Must supply db configuration')
        self._dbconfig = dbconfig
        self._db = None
        self.initialize_database()
    
    """
        Convenience method to return db, calling db_open() if needed
    """    
    @property
    def db(self):
        return self._db or self.db_open()
    
    """
        If database file doesn't exist:
            create it
        If tables don't exist:
            create them
    """
    def initialize_database(self):
        # Where does the user want the database to live?
        name = self._dbconfig['name']
        # turn into an absolute path:
        name = os.path.abspath(name)
        # save absolute path back into config so
        # that other parts of the program can use it.
        self._dbconfig['name'] = name
        # If directory for db doesn't already exist,
        # create it. Use os.makedirs since this does
        # a recursive, multi-level mkdir if needed.
        dbdir = os.path.dirname(name)
        print 'will create db file in {}'.format(dbdir)
        if not os.path.isdir(dbdir):
            os.makedirs(dbdir)
        # Set up db file and tables:
        self.execute_sql_resource('data' + os.sep + 'pogo_schema.sql')
    
    
    """
        Open the named resource, which should be a list
        of sql commands, separated by semicolons. Execute
        each statement.
    """
    def execute_sql_resource(self, resource_name):
        data = resource_string('pogo', resource_name)
        data.replace(os.linesep, '') # strip newlines
        data = data.strip() # and leading/trailing whitespace
        commands = data.split(';') # split on SQL end-of-command marker
        cursor = self.db.cursor()
        cursor.execute('BEGIN TRANSACTION')
        for c in commands:
            cursor.execute(c)
        cursor.execute('COMMIT')

    def db_open(self):
        if self._db is None:
            if (self._dbconfig['type'] == 'sqlite'):
                self._db = sqlite3.connect(self._dbconfig['name'])  # @UndefinedVariable
                self._db.isolation_level = None # Do this to turn off automatic transactions.
            else:
                raise ValueError('Unsupported database type')
        return self._db
    
    def db_close(self):
        if self._db is not None:
            self._db.close()
            self._db = None
    


import iso8601
import time
from datetime import datetime
from tzlocal import get_localzone
from geoip import geolite2
import logging
import os.path
import sqlite3
import sys
import tarfile
        

# Convert timestamp in local time to GMT:
# Requires modules iso8601 and time
def local_timestamp_to_gmt(localtimestamp):
    if not localtimestamp:
        return None
    d = iso8601.parse_date(localtimestamp)
    utc_seconds = time.mktime(d.timetuple())
    return time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(utc_seconds))


"""
    We're given a string representing a date and time.
    We know it's in US Eastern Time, but don't know whether
    daylight savings time was in effect at that time. In
    other words, we know the offset from UTC is either -5 hours
    or - 4 hours, but we don't know which. We'd like to
    convert that date/time to UTC.
"""
def local_no_tz_to_utc(datetime_string):
    if not datetime_string:
        return None
    lnaive = datetime.strptime(datetime_string, "%Y-%m-%d %H:%M:%S")
    tz=get_localzone()
    l_with_tz = tz.localize(lnaive)
    #l_with_tz now has timezone info
    utc_seconds = time.mktime(l_with_tz.timetuple())
    return time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(utc_seconds))

"""
    Given a logging level string such as 'DEBUG' or 'INFO',
    get the corresponding numeric logging level.
"""
def logging_level_from_string(levelstr):
    numeric_level = getattr(logging, levelstr.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % levelstr)
    return numeric_level

"""
    Use the information in the app's configuration
    to set logging options.
"""
def configure_logging(loginfo):
    fn=loginfo()['filename']
    levelstr=loginfo()['level']
    if levelstr == '': levelstr = 'NOTSET'
    level = logging_level_from_string(levelstr)
    if fn != 'CONSOLE':
        logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', filename=fn, level=level)
    else:
        logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=level)
    return logging.getLogger()

"""
    Generate an archive name
    based on the passed-in prefix and the current date and time.
"""
def generate_archive_name(prefix = None):
    utcnow = datetime.utcnow()
    if prefix is None:
        prefix = ''
    return prefix + utcnow.isoformat() + '.bz2'

"""
    Given a list of files and an archive file name,
    create the archive file and add the files from the list.
    Try to figure out what kind of compression to use
    based on the archive file extension.
"""
def archive_file_list(filename, files):
    comp_mode = ''
    if filename.endswith('bz2'): comp_mode = ':bz2'
    if filename.endswith('gz'): comp_mode = ':gz'
    open_mode_string = 'w' + comp_mode
    with tarfile.open(filename, open_mode_string) as tf:
        for fn in files:
            tf.add(fn)


class PogoGeoInfo(object):
    def __init__(self, ipi=None): # ipi is a geoip.IPInfo
        if ipi is None:
            self.country_name = ''
            self.country_code = ''
        else:
            self._info_dict = ipi.get_info_dict()
            # several pieces of data in _info_dict are used almost
            # every time this class is used, so save them in an
            # easily-accessible place:
            # Other items are still accessible by dereferencing _info_dict.
            if not self._info_dict['country']:
                self.country_name = ''
                self.country_code = ''
            else:
                if self._info_dict['country']['names'] and self._info_dict['country']['names']['en']:
                    self.country_name = self._info_dict['country']['names']['en']
                else:
                    self.country_name = ''
                if self._info_dict['country']['iso_code']:
                    self.country_code = self._info_dict['country']['iso_code']
                else:
                    self.country_code = ''

"""
    Return an object holding information about
    location of given ip address. If no ip address
    is given, instantiate a PgooGeoInfo object with None,
    which will set country_code and country_name to ''.
"""
def get_geo_info(ipaddress):
    if not ipaddress:
        return PogoGeoInfo()
    else:
        return PogoGeoInfo(geolite2.lookup(ipaddress))
    
     


"""
    A StretchFile reads disk files into RAM and creates Record objects from their contents.
"""
import abc
import os
import os.path
import re

from pogo.dto.record import LogRecord, AttemptRecord, SessionLogRecord, SessionDownloadFileRecord
from pogo.dto.record import SessionRecordingRecord
from pogo.util.util import get_geo_info

class StretchFile(object):
    __metaclass__ = abc.ABCMeta
    def __init__(self, file_name):
        self._name = file_name
        self._entry_list = []
        self._loaded = False
    
    def name(self):
        return self._name
    
    def __iter__(self):
        return iter(self._entry_list)
    
    # override __len__ so that len(bungeedatafile)
    # returns the number of files in entry_list
    def __len__(self):
        return len(self._entry_list)
    
    @abc.abstractmethod
    def load(self):
        pass


class AttemptFile(StretchFile):
    def __init__(self, file_name):
        super(AttemptFile, self).__init__(file_name)

    def load(self):
        if os.path.isfile(self.name()):
            try:
                f = open(self.name(), "rt")
                try:
                    for line in f:
                        r = AttemptRecord(line.rstrip())
                        self._entry_list.append(r)
                    f.close()
                    self._loaded = True
                    return True
                except IOError:
                    if f is not None:
                        f.close()
                    logging.error("Failed to load file ", exc_info = True)
                    print "During loading of " + self.name() + " encountered i/o error reading line " + line
                    self._loaded = False
                    return False
            except IOError:
                logging.error("Could not open file", exc_info = True)
                self._loaded = False
                return False
        else: 
            raise ValueError(self.name() + " is not a file.")
            self._loaded = False
            return False
    

class LogFile(StretchFile):
    EXTRA_LINE_PATTERN = re.compile('^\t')

    def __init__(self, file_name):
        super(LogFile, self).__init__(file_name)
        
    def load(self):
        if os.path.isfile(self.name()):
            try:
                f = open(self.name(), "rt")
                try:
                    for line in f:
                        if LogFile.EXTRA_LINE_PATTERN.match(line):
                            # Don't use this to construct a
                            # LogRecord. Instead, tack it on to
                            # the end of the last entry's message
                            # field.
                            if len(self._entry_list) > 0:
                                r = self._entry_list.pop()
                                r.message += ' -- ' + line.strip()
                        else:
                            r = LogRecord(line.rstrip()) 
                        self._entry_list.append(r)
                        
                    f.close()
                    self._loaded = True
                    return True
                except IOError:
                    if f is not None:
                        f.close()
                    logging.error("Could not read a line in file", exc_info = True)    
                    print "During loading of " + self.name() + " encountered i/o error reading line " + line
                    self._loaded = False
                    return False
            except IOError:
                logging.error("Failed to load file ", exc_info = True)
                print "Error during loading of file " + self.name()
                self._loaded = False
                return False
        else: 
            raise ValueError(self.name() + " is not a file.")
            self._loaded = False
            return False
# end of LogFile.load())


class SessionLogFile(StretchFile):
    def __init__(self, file_name):
        super(SessionLogFile, self).__init__(file_name)
        # Get second-to-last element of filespec -- this is the IP address
        source_ip = file_name.split(os.sep)[-2]
        self.set_source_ip(source_ip)
        
    def set_source_ip(self, source_ip):
        if source_ip:
            self.source_ip = source_ip
            gpi = get_geo_info(self.source_ip)
            self.country_code = gpi.country_code
            self.country_name = gpi.country_name
        else:
            self.source_ip = ''
            self.country_code = ''
            self.country_name = ''

        
    def load(self):
        if os.path.isfile(self.name()):
            try:
                f = open(self.name(), "rt")
                try:
                    for line in f:
                        r = SessionLogRecord(line)
                        r.set_source_ip(self.source_ip)
                        r.set_country_info(self.country_code, self.country_name)
                        self._entry_list.append(r)
                    f.close()
                    self._loaded = True
                    return True
                except IOError:
                    if f is not None:
                        f.close()
                        logging.error("Failed to load file ", exc_info = True)
                        print "During loading of " + self.name() + " encountered i/o error reading line " + line
                    self._loaded = False
                    return False
            except IOError:
                logging.error("Failed to load file ", exc_info = True)
                print "Error during loading of file " + self.name()
                self._loaded = False
                return False
        else: 
            raise ValueError(self.name() + " is not a file.")
            self._loaded = False
            return False
# end of SessionLogFile.load())


class SessionDownloadFile(StretchFile):
    def __init__(self, file_name):
        super(SessionDownloadFile, self).__init__(file_name)
        # IP is third-to-last element of filespec
        source_ip = file_name.split(os.sep)[-3]
        self.set_source_ip(source_ip)
        
    def set_source_ip(self, source_ip):
        if source_ip:
            self.source_ip = source_ip
            gpi = get_geo_info(self.source_ip)
            self.country_code = gpi.country_code
            self.country_name = gpi.country_name
        else:
            self.source_ip = ''
            self.country_code = ''
            self.country_name = ''

        
    def load(self):
        if os.path.isfile(self.name()):
            try:
                with open(self.name(), "rb") as f:
                    data = f.read()
                    if data:
                        data64 = data.encode("base64")
                    else:
                        data64 = ''
                    r = SessionDownloadFileRecord(self.name(), data64)
                    # Part of file name is a datetime stamp. Extract it
                    namepart = self.name().split(os.sep)[-1] # get last element of filespec
                    # datetime string in format expected by set_timestamp():
                    normalized_namepart = ( namepart[0:4] + '-'
                                             + namepart[4:6] + '-'
                                             + namepart[6:8] + ' '
                                             + namepart[9:11] + ':'
                                             + namepart[11:13] + ':'
                                             + namepart[13:15] )
                    r.set_timestamp(normalized_namepart)
                    r.set_source_ip(self.source_ip)
                    r.set_country_info(self.country_code, self.country_name)
                    self._entry_list.append(r)
                    self._loaded = True
                    return True
            except IOError:
                logging.error("Failed to load file ", exc_info = True)
                print "During loading of " + self.name() + " encountered i/o error"
                self._loaded = False
                return False
        else: 
            raise ValueError(self.name() + " is not a file.")
            self._loaded = False
            return False
# end of SessionDownloadFile.load())


class SessionRecordingFile(StretchFile):
    def __init__(self, file_name):
        super(SessionRecordingFile, self).__init__(file_name)
        # IP is second-to-last element of filespec
        source_ip = file_name.split(os.sep)[-2]
        self.set_source_ip(source_ip)
        
    def set_source_ip(self, source_ip):
        if source_ip:
            self.source_ip = source_ip
            gpi = get_geo_info(self.source_ip)
            self.country_code = gpi.country_code
            self.country_name = gpi.country_name
        else:
            self.source_ip = ''
            self.country_code = ''
            self.country_name = ''

        
    def load(self):
        if os.path.isfile(self.name()):
            try:
                with open(self.name(), "rb") as f:
                    data = f.read()
                    if data:
                        data64 = data.encode("base64")
                    else:
                        data64 = ''
                    r = SessionRecordingRecord(self.name(), data64)
                    # Part of file name is a datetime stamp. Extract it
                    namepart = self.name().split(os.sep)[-1] # get last element of filespec
                    # datetime string in format expected by set_timestamp():
                    normalized_namepart = ( namepart[0:4] + '-'
                                             + namepart[4:6] + '-'
                                             + namepart[6:8] + ' '
                                             + namepart[9:11] + ':'
                                             + namepart[11:13] + ':'
                                             + namepart[13:15] )
                    r.set_timestamp(normalized_namepart)
                    r.set_source_ip(self.source_ip)
                    r.set_country_info(self.country_code, self.country_name)
                    self._entry_list.append(r)
                    self._loaded = True
                    return True
            except IOError:
                logging.error("Failed to load file ", exc_info = True)
                print "During loading of " + self.name() + " encountered i/o error"
                self._loaded = False
                return False
        else: 
            raise ValueError(self.name() + " is not a file.")
            self._loaded = False
            return False
# end of SessionRecordingFile.load())






"""
    Classes in this file are responsible for writing to and reading from
    the ElasticSearch database on the log server.
"""
import abc
#from elasticutils import get_es
from elasticsearch import Elasticsearch

class RecordDaoES(object):
    __metaclass__ = abc.ABCMeta
    def __init__(self, es_cfg):
        self._make_es_connection(es_cfg)
        self._assure_index()
        self._assure_mapping()

    def _make_es_connection(self, es_cfg):
        if es_cfg is None:
            raise ValueError("RecordDaoEs needs ElasticSearch Configuration Information")
        else:
            self._es_index = es_cfg['es_index']
            self._es_host = es_cfg['es_host']
            self._es_port = es_cfg['es_port']
            self._es_timeout = es_cfg['es_timeout']
            if not self._es_timeout: self._es_timeout = 30
			# needed to work around problems in some versions of urllib3:
            port_as_float = float(self._es_port)
            timeout_as_float = float(self._es_timeout)

            es = elasticsearch.Elasticsearch([{u'host': self._es_host, u'port': port_as_float, u'timeout': timeout_as_float }])
            if not es: raise Exception("Could not initialize Elasticsearch connection.")
            self._es_connection = es
            #~ # Open the connection to the ES server and make sure our index exists
            #~ es_url = 'http://' + self._es_host
            #~ if self._es_port:
                #~ es_url += ':' + self._es_port
            #~ self._es_connection = get_es(urls=[es_url], timeout=self._es_timeout)
            #~ if not self._es_connection:
                #~ raise Exception("Could not initialize ElasticSearch connection.")

    def _assure_index(self):
        # Make sure our index exists:
        if not self._es_connection.indices.exists(index=self._es_index):
            self._es_connection.indices.create(index=self._es_index)


    def _assure_mapping(self):
        # Verify our mapping exists.
        es_res = self._es_connection.indices.get_mapping(index=self._es_index)
        # If es_res does not have key INDEX_NAME then the mappings don't exist
        # either.
        if (not es_res.has_key(self._es_index) or
            not es_res[self._es_index].has_key('mappings') or
            not es_res[self._es_index]['mappings'].has_key(self.get_document_type())
            ):
            self._es_connection.indices.put_mapping(self.get_document_type(),
                 {'properties': self.get_mapping() }, self._es_index)


    def insert_single(self, record):
        d = record.as_dict()
        t = self.get_document_type()
        idx = self._es_index
        r = self._es_connection.index(index=idx, doc_type=t, body=d)
        return r['_id']

    abc.abstractmethod
    def get_document_type(self):
        return ''

    abc.abstractmethod
    def get_mapping(self):
        return ''


class AttemptRecordDaoES(RecordDaoES):
    DOCUMENT_TYPE = 'HonSSH_Attempt'
    MAPPING = {
        "bifrozt_host": {"type": "string", "index": "not_analyzed" },
        "timestamp": {"type": "date", "format": "YYYY-MM-dd HH:mm:SS"},
        "source_ip": {"type": "string", "index": "not_analyzed" },
        "country_name": {"type": "string", "index" : "not_analyzed" },
        "country_code": {"type": "string", "index" : "not_analyzed" },
        "user": {"type" : "string", "index" : "not_analyzed"},
        "password": {"type": "string", "index": "not_analyzed"},
        "success": {"type": "boolean"}
    }

    def __init__(self, es_cfg):
        super(AttemptRecordDaoES, self).__init__(es_cfg)

    def get_document_type(self):
        return AttemptRecordDaoES.DOCUMENT_TYPE

    def get_mapping(self):
        return AttemptRecordDaoES.MAPPING



class LogRecordDaoES(RecordDaoES):
    DOCUMENT_TYPE = 'HonSSH_LogEntry'
    MAPPING = {
           "bifrozt_host": {  "type": "string", "index": "not_analyzed" },
          "timestamp": {"type": "date", "format": "YYYY-MM-dd HH:mm:SS"},
           "server_info": {"type": "string", "index": "not_analyzed"},
           "message": {"type": "string"}
           }

    def __init__(self, es_cfg):
        super(LogRecordDaoES, self).__init__(es_cfg)


    def get_document_type(self):
        return LogRecordDaoES.DOCUMENT_TYPE

    def get_mapping(self):
        return LogRecordDaoES.MAPPING

class SessionLogDaoES(RecordDaoES):
    DOCUMENT_TYPE = 'HonSSH_SessionLogEntry'
    MAPPING = {
           "bifrozt_host": {  "type": "string", "index": "not_analyzed" },
           "timestamp": {"type": "date", "format": "YYYY-MM-dd HH:mm:SS"},
           "source_ip": { "type": "string", "index": "not_analyzed" },
           "country_name": { "type": "string", "index" : "not_analyzed" },
           "country_code": { "type": "string", "index" : "not_analyzed" },
           "channel": {"type": "string", "index": "not_analyzed"},
           "message": {"type": "string"}
           }

    def __init__(self, es_cfg):
        super(SessionLogDaoES, self).__init__(es_cfg)

    def get_document_type(self):
        return SessionLogDaoES.DOCUMENT_TYPE

    def get_mapping(self):
        return SessionLogDaoES.MAPPING


class SessionRecordingDaoES(RecordDaoES):
    DOCUMENT_TYPE = 'HonSSH_SessionRecording'
    MAPPING = {
       "bifrozt_host": {  "type": "string", "index": "not_analyzed" },
       "timestamp": {"type": "date", "format": "YYYY-MM-dd HH:mm:SS"},
       "source_ip": { "type": "string", "index": "not_analyzed" },
       "country_name": { "type": "string", "index" : "not_analyzed" },
       "country_code": { "type": "string", "index" : "not_analyzed" },
       "filename": {"type": "string", "index": "not_analyzed"},
       "contents": {"type": "string", "index": "no"}
       }

    def __init__(self, es_cfg):
        super(SessionRecordingDaoES, self).__init__(es_cfg)

    def get_document_type(self):
        return SessionRecordingDaoES.DOCUMENT_TYPE

    def get_mapping(self):
        return SessionRecordingDaoES.MAPPING


class SessionDownloadDaoES(RecordDaoES):
    DOCUMENT_TYPE = 'HonSSH_SessionDownload'
    MAPPING = {
       "bifrozt_host": {  "type": "string", "index": "not_analyzed" },
       "timestamp": {"type": "date", "format": "YYYY-MM-dd HH:mm:SS"},
       "source_ip": { "type": "string", "index": "not_analyzed" },
       "country_name": { "type": "string", "index" : "not_analyzed" },
       "country_code": { "type": "string", "index" : "not_analyzed" },
       "filename": {"type": "string", "index": "not_analyzed"},
       "contents": {"type": "string", "index": "no"}
       }

    def __init__(self, es_cfg):
        super(SessionDownloadDaoES, self).__init__(es_cfg)

    def get_document_type(self):
        return SessionDownloadDaoES.DOCUMENT_TYPE

    def get_mapping(self):
        return SessionDownloadDaoES.MAPPING







"""
    Process HonSSH data -- get it from the files generated by HonSSH
    into the ElasticSearch database. There are three steps:
    
    1. "Scraping." In this step, the files are read and their contents
    are stored in a database that's local to the HonSSH process. Currently,
    sqlite is used; other database systems would be feasible if needed. After
    a file's content is scraped, the file is marked as done.
    
    2. Inserting into ElasticSearch. In this step, the database records are
    read and used as the input to ElasticSearch insertion methods. Records that are
    successfully inserted into ElasticSearch are marked as such by having their
    ElasticSearch-assigned id strings stored in the local database.
    
    3. Cleanup. In this step, files and database records marked as done in steps
    1 and 2 are deleted.
"""
import sqlite3
import logging
import sys
import os

from pogo.dao.record_dao_es import AttemptRecordDaoES, LogRecordDaoES
from pogo.dao.record_dao_es import SessionLogDaoES, SessionRecordingDaoES, SessionDownloadDaoES
from pogo.dao.record_dao_local import AttemptRecordDaoLocal, LogRecordDaoLocal
from pogo.dao.record_dao_local import SessionLogDaoLocal, SessionRecordingDaoLocal, SessionDownloadDaoLocal
from pogo.dao.local_db_access import LocalDBAccessor
from pogo.dto.record import AttemptRecord, LogRecord
from pogo.dto.record import SessionLogRecord, SessionRecordingRecord, SessionDownloadFileRecord
from pogo.file.file_lister import AttemptFileLister, LogFileLister
from pogo.file.file_lister import SessionLogFileLister, SessionRecordingFileLister, SessionDownloadFileLister
from pogo.service.service_local import ServiceLocal
from pogo.util.config import StretchConfig
from pogo.util.util import logging_level_from_string, configure_logging
from pogo.util.util import generate_archive_name, archive_file_list
import os.path


class Pogo(object ):
    def __init__(self):
        self._cfg = StretchConfig()
        self._logger = configure_logging(self._cfg.get_logging_info)
        self._dba = LocalDBAccessor(self._cfg.get_db_info())
        self._dba.initialize_database()
        # Create directory to store archived data files, if it doesn't already exist:
        self._arc_dir = self._cfg.get_locations()['arc_dir']
        if not os.path.isdir(self._arc_dir):
            os.makedirs(self._arc_dir)
        
    def scrape_honssh_files(self, loc_type, lister_class, dao_local_class):
        source_dir = self._cfg.get_locations()[loc_type]
        lister = lister_class(source_dir)
        lister.load_file_name_lists()
        lister.load_pending_file_objects()
        self._logger.info("File lister loaded with %s files", len(lister))
        print "File lister loaded with " + str(len(lister)) + " files"
        dao_obj = dao_local_class(self._dba)
        aservice = ServiceLocal(dao_obj)
        total_num_saved = 0
        done_files = []
        for f in lister:
            if f.load():
                self._logger.info("loaded %s, containing %s records", f.name(), len(f) )
                print "loaded " + f.name() + " containing " + str(len(f)) + " records"
                if len(f) > 0:
                    num_saved = aservice.write_new_records(f._entry_list)
                    self._logger.info("Saved %s records", num_saved)
                    print "Number saved: " + str(num_saved)
                    total_num_saved += num_saved
                    if num_saved is None or num_saved != len(f):
                        raise Exception("Only " + num_saved + "records written from file " + f._name + ". File contains " + len(f) + " + records.")
                    else:
                        lister.mark_as_done(f)
                        done_files.append(f._name)
        return done_files
    
    
    def scrape_session_log_records(self):
        return self.scrape_honssh_files('session_dir', SessionLogFileLister, SessionLogDaoLocal)
    
    def scrape_session_download_files(self,):
        return self.scrape_honssh_files('session_dir', SessionDownloadFileLister, SessionDownloadDaoLocal)
    
    def scrape_session_recordings(self,):
        return self.scrape_honssh_files('session_dir', SessionRecordingFileLister, SessionRecordingDaoLocal)
    
    def scrape_attempt_records(self,):
        return self.scrape_honssh_files('attempt_dir', AttemptFileLister, AttemptRecordDaoLocal)
    
    def scrape_log_records(self,):
        return self.scrape_honssh_files('log_dir', LogFileLister, LogRecordDaoLocal)
    
    def put_records_into_es(self, localdaoclass, esclass, recordclass):
            db_local = localdaoclass(self._dba)
            aservice = ServiceLocal(db_local)
            es_link = esclass(self._cfg.get_es_info())
            rows = aservice.get_non_processed()
            total_to_add = len(rows)
            self._logger.info("Found %s records not yet put into ES", total_to_add)
            print "Found " + str(total_to_add) + " records not yet put into ES"
            num_into_es = 0
            for row in rows:
                r = recordclass()
                i = 0
                for fld in db_local.get_insert_fields():
                    setattr(r, fld, row[i+2])
                    i += 1
                es_id = es_link.insert_single(r)
                if es_id:
                    try:
                        db_id = row[0]
                        aservice.update_with_es_id(db_id, es_id)
                    except sqlite3.Error as e:
                        self._logger.error("Could not update record in local db", exc_info = True)
                        print "Could not update record with database id " + str(row[0]) + " in local database!"
                        raise e
                    num_into_es += 1
                    # emit progress indication...
                    if (self._logger.isEnabledFor(logging.INFO)):
                        if num_into_es % 50 == 0:
                            self._logger.info("Added to ES: %s of %s...", num_into_es, total_to_add)
                else:
                    raise Exception("Could not add record with database id " + str(row[0]) + " to ElasticSearch!")
            return num_into_es
        
    def put_session_log_records_into_es(self):
        return self.put_records_into_es(SessionLogDaoLocal, SessionLogDaoES, SessionLogRecord)
        
    def put_session_download_records_into_es(self):
        return self.put_records_into_es(SessionDownloadDaoLocal, SessionDownloadDaoES, SessionDownloadFileRecord)
        
    def put_session_recordings_into_es(self):
        return self.put_records_into_es(SessionRecordingDaoLocal, SessionRecordingDaoES, SessionRecordingRecord)
    
    def put_attempt_records_into_es(self):
        return self.put_records_into_es(AttemptRecordDaoLocal, AttemptRecordDaoES, AttemptRecord)
    
    def put_log_records_into_es(self):
        return self.put_records_into_es(LogRecordDaoLocal, LogRecordDaoES, LogRecord)
    
    
    def prune_honssh_records(self, loc_type, lister_class, dao_local_class):
        source_dir = self._cfg.get_locations()[loc_type]
        lister = lister_class(source_dir)
        lister.load_file_name_lists()
        self._logger.info("Found %s files to prune", len(lister._done_file_names) )
        print "Found {0} files to prune".format(len(lister._done_file_names))
        count_files_removed = lister.delete_done_files()
        self._logger.info("Removed %s files", count_files_removed)
        print "Removed {0} files.".format(count_files_removed)
        self._logger.info('Will now attempt to remove processed records from local database...')
        print "Will now attempt to remove processed records from local database..."
        db_local = dao_local_class(self._dba)
        aservice = ServiceLocal(db_local)
        count_db_rows_deleted = aservice.delete_finished_records()
        self._logger.info("Removed %s records from local database", count_db_rows_deleted)
        return (count_files_removed, count_db_rows_deleted)
    
    
        
    def prune_session_log_records(self):
        return self.prune_honssh_records('session_dir', SessionLogFileLister, SessionLogDaoLocal)
    
    def prune_session_download_records(self):
        return self.prune_honssh_records('session_dir', SessionDownloadFileLister, SessionDownloadDaoLocal)
    
    def prune_session_recordings(self):
        return self.prune_honssh_records('session_dir', SessionRecordingFileLister, SessionRecordingDaoLocal)
    
    def prune_attempt_records(self):
        return self.prune_honssh_records('attempt_dir', AttemptFileLister, AttemptRecordDaoLocal)
    
    def prune_log_records(self):
        return self.prune_honssh_records('log_dir', LogFileLister, LogRecordDaoLocal)
    
    def main(self):
        td = self._arc_dir
        files_scraped = self.scrape_attempt_records()
        if len(files_scraped) > 0:
            arc_name = generate_archive_name(td + os.sep + 'HonSSH_Attempts-')
            archive_file_list(arc_name, files_scraped)
#         self.scrape_log_records()
        files_scraped = self.scrape_session_download_files()
        if len(files_scraped) > 0:
            arc_name = generate_archive_name(td + os.sep + 'HonSSH_Session_Downloads-')
            archive_file_list(arc_name, files_scraped)

        files_scraped = self.scrape_session_log_records()
        if len(files_scraped) > 0:
            arc_name = generate_archive_name(td + os.sep + 'HonSSH_Session_Logs-')
            archive_file_list(arc_name, files_scraped)

        files_scraped = self.scrape_session_recordings()
        if len(files_scraped) > 0:
            arc_name = generate_archive_name(td + os.sep + 'HonSSH_Session_Recordings-')
            archive_file_list(arc_name, files_scraped)
        
        self.put_attempt_records_into_es()
        self.put_log_records_into_es()
        self.put_session_download_records_into_es()
        self.put_session_log_records_into_es()
        self.put_session_recordings_into_es()
        
        self.prune_attempt_records()
        self.prune_log_records()
        self.prune_session_download_records()
        self.prune_session_log_records()
        self.prune_session_recordings()


def main():
    b = Pogo()
    b.main()
        
if __name__ == '__main__':
    main()


import ConfigParser
import os
import os.path
import sys

class StretchConfig(object):
    
    def __init__(self):
        def_top_dir = '/opt/honssh'
        def_db_dir = '/usr/local/share/pogo/db'
        config_file_search_path = [ '/etc/pogo.cfg', '/etc/default/pogo.cfg', '/etc/boing/pogo.cfg',
                        '/usr/local/share/pogo/pogo.cfg',  '~/.config/boing/pogo.cfg', './pogo.cfg']

        # Default settings:
        self._settings = {
                        'debug': 0,
                        'locations': {
                                      'top_dir': def_top_dir,
                                      'log_dir': def_top_dir + os.sep + 'logs',
                                      'session_dir': def_top_dir + os.sep + 'sessions',
                                      'attempt_dir': def_top_dir + os.sep + 'logs',
                                      'arc_dir': def_top_dir + os.sep + 'archives'
                                      },
                        'elasticsearch': {
                                          'es_host': 'localhost',
                                          'es_port': '9200',
                                          'hon_index': 'hon_ssh'                  
                                          },
                        'db_connection': {
                                          'type': 'sqlite',
                                          'host': '',
                                          'port': '',
                                          'user': '',
                                          'password': '',
                                          'name': def_db_dir + os.sep + 'pogo.db'
                                          },
                          'logging': {
                                      'filename': 'CONSOLE',
                                      'level': 'WARNING'
                                      }
                        }
        
        # See if we can read a config file:
        cfg = ConfigParser.SafeConfigParser()
        self.config_files_read = cfg.read(config_file_search_path)

        
        # If we found a config file, override
        # defaults with values from config file:
        # special case for debug, since it's a boolean (rest are strings)
        if cfg.has_section('main'):
            self._settings['debug'] = cfg.getboolean('main', 'debug')
  
        for section in ('locations', 'db_connection', 'elasticsearch', 'logging'):
            if cfg.has_section(section):
                for item in cfg.items(section):
                    self._settings[section][item[0]] = item[1]
#         
            

    def __str__(self, *args, **kwargs):
        retStr = 'StretchConfig: \n\tDebug: ' + str(self._settings['debug']) + '\n'
        for section in ('locations', 'db_connection', 'elasticsearch', 'logging'):
            retStr += '\t' + section + ' section:\n'
            for key in self._settings[section]:
                retStr += '\t\t' + key + ': ' + self._settings[section][key] + '\n'
        return retStr
    
    def get_db_info(self):
        return self._settings['db_connection']
    
    def get_locations(self):
        return self._settings['locations']
    
    def get_es_info(self):
        return self._settings['elasticsearch']
    
    def get_logging_info(self):
        return self._settings['logging']
            

if __name__ == '__main__':
    bc = StretchConfig()
    if bc.config_files_read is not None and len(bc.config_files_read) > 0:
        print "Read one or more config files: "
        for f in bc.config_files_read: print f
    else:
        print "No config files read; using default values"
    
    print bc


"""
    dto.Record is an abstract base class. Its subclasses hold
    information about individual data records, which are
    represented by individual lines in the HonSSH-generated files,
    individual rows in the local db storage tables, or individual
    "documents" in the ElasticSearch database.
"""
import abc
from pogo.util.util import local_timestamp_to_gmt, local_no_tz_to_utc, get_geo_info

from socket import gethostname
class Record(object):
    def __init__(self):
        __metaclass__ = abc.ABCMeta
        self.bifrozt_host = gethostname()
        self.db_id = ''
        self.es_id = ''
    @abc.abstractmethod    
    def as_dict(self):
        raise Exception('Abstract methods should not be called.')
    
    # Comparison operator method
    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    
    def __ne__(self, other):
        return not (self == other)
    # end of base class
    
class AttemptRecord(Record):
    def __init__(self, logLine=None):
        super( AttemptRecord, self ).__init__()
        
        # If we were passed a string, use that to
        # initialize ourself:
        if logLine is not None:
            splitList = logLine.split(',')
            # if < 5 elements in splitList, append '' to bring length to 5
            while (len(splitList) < 5):
                splitList.append('')
            timestring = splitList[0]
            # We know that timestring is in Eastern time, but
            # don't know if daylight savings time was in effect or
            # not. Use local_no_tz_to_utc() to convert to UTC.
            self.timestamp = local_no_tz_to_utc(timestring)
            self.source_ip = splitList[1]
            # Try to protect against invalid chars
            # in user-entered values:
            self.user = unicode(splitList[2], errors='replace')
            self.password = unicode(splitList[3], errors='replace')
            self.success = splitList[4]
            if not self.success: self.success = '0'
            gpi = get_geo_info(self.source_ip)
            self.country_code = gpi.country_code
            self.country_name = gpi.country_name
        else:
            # initialize fields to empty strings or null values:
            self.timestamp = ''
            self.source_ip = ''
            self.user = ''
            self.password = ''
            self.success = '0' # assume failure
            self.country_code = ''
            self.country_name = ''
    
    @abc.abstractmethod
    def as_dict(self):
        # make a structure to hold our stuff
        adict = {}
        if self.timestamp:
            # if timestamp exists and isn't blank...
            adict['bifrozt_host'] = self.bifrozt_host
            adict['source_ip'] = self.source_ip
            adict['user'] = self.user
            adict['password'] = self.password
            adict['success'] = self.success
            adict['timestamp'] = self.timestamp
            adict['country_name'] = self.country_name
            adict['country_code'] = self.country_code
        return adict   
    
    # end of AttemptRecord class  
    
class LogRecord(Record):
    def __init__(self, log_line = None):
        super( LogRecord, self ).__init__()
        if log_line is not None:
            splitList = log_line.split(' ')
            local_time_string = splitList[0] + ' ' + splitList[1]
            self.timestamp = local_timestamp_to_gmt(local_time_string)
            self.server_info = splitList[2]
            self.message = splitList[3]
        else:
            # initialize to empty fields
            self.timestamp = ''
            self.server_info = ''
            self.message = ''
        
    @abc.abstractmethod    
    def as_dict(self):
        # make a structure to hold our stuff
        adict = {}
        if self.timestamp:
            # if timestamp exists and isn't blank...
            adict['bifrozt_host'] = self.bifrozt_host
            adict['timestamp'] = self.timestamp
            adict['server_info'] = self.server_info
            adict['message'] = self.message
        return adict





class SessionRecord(Record):
    def __init__(self):
        super( SessionRecord, self ).__init__()

    def as_dict(self):
        # make a structure to hold our stuff
        adict = {}
        if self.bifrozt_host:
            adict['bifrozt_host'] = self.bifrozt_host
            adict['timestamp'] = self.timestamp
            adict['source_ip'] = self.source_ip
            adict['country_name'] = self.country_name
            adict['country_code'] = self.country_code
        return adict   

    def set_source_ip(self, source_ip):
        if source_ip:
            self.source_ip = source_ip
        else:
            self.source_ip = ''
    
    def set_country_info(self, code, name):
        if code:
            self.country_code = code
        else:
            self.country_code = ''
        if name:
            self.country_name = name
        else:
            self.country_name = ''
            
    def set_timestamp(self, timestring):
        if timestring:
            self.timestamp = local_no_tz_to_utc(timestring)
        else:
            self.timestamp = ''
        

class SessionLogRecord(SessionRecord):
    def __init__(self, line = None):
        super( SessionLogRecord, self ).__init__()
        if line :
            line = line.strip()
            if len(line) > 30:
                self.set_timestamp(line[0:19])
                lb_ind = line.index('[')
                rb_ind = line.index(']')
                chan = line[lb_ind+1:rb_ind] # channel
                self.channel = chan.strip()
                # Now take everything after the last ']':
                self.message = line[rb_ind+1:] # rest of the line
            else:
                self.set_timestamp('')
                self.set_source_ip('')
                self.set_country_info('', '')
                self.channel = ''
                self.message = ''
        else:
            self.set_timestamp('')
            self.set_source_ip('')
            self.set_country_info('', '')
            self.channel = ''
            self.message = ''
    
    def as_dict(self):
        # make a structure to hold our stuff
        adict = super(SessionLogRecord,self).as_dict()
        adict['channel'] = self.channel
        adict['message'] = self.message
        return adict   


class SessionRecordingRecord(SessionRecord):
    def __init__(self, filename=None, contents=None):
        super( SessionRecordingRecord, self ).__init__()
        self.set_contents(contents)
        self.set_filename(filename)
            
    def set_filename(self, name):
        if name is not None:
            self.filename = name
        else:
            self.filename = ''
    
    def set_contents(self, contents):
        if contents is not None:
            self.contents = contents
        else:
            self.contents = ''
    
    def as_dict(self):
        # make a structure to hold our stuff
        adict = super(SessionRecordingRecord,self).as_dict()
        adict['filename'] = self.filename
        adict['contents'] = self.contents
        return adict   

class SessionDownloadFileRecord(SessionRecord):
    def __init__(self, fname=None, contents=None):
        super( SessionDownloadFileRecord, self ).__init__()
        self.set_contents(contents)
        self.set_filename(fname)
            
    def set_filename(self, name):
        if name is not None:
            self.filename = name
        else:
            self.filename = ''
    
    def set_contents(self, contents):
        if contents is not None:
            self.contents = contents
        else:
            self.contents = ''
    
    def as_dict(self):
        # make a structure to hold our stuff
        adict = super(SessionDownloadFileRecord,self).as_dict()
        adict['filename'] = self.filename
        adict['contents'] = self.contents
        return adict   




"""
    While the Log and Attempt record types are collections of strings, a Session is
    not. HonSSH generates the following kinds of session information:
        1. session logs. These are simple text files, with each line listing:
            A. a timestamp
            B. a "channel," (ie, whether this event involved communication via an SSH connection,
            typing at a terminal command line, or something else.
            C. a description, or the text of the command executed
            
        2. recording files. These are binary files in a format suitable for playback with
        the playback.py utility included with HonSSH.
        
        3. timestamped copies of any files downloaded by users during SSH sessions.
        
    The session information can be associated with an originating IP because HonSSH stores
    all of this in a directory with that IP as its name.
    
    However, it's difficult to associate a given file, for example, with a particular session log or
    session recording -- their dates and times may or may not be the same, depending on
    the exact sequence and timing of events. Therefore, this program will treat the types of
    information as separate entities rather than as part of a single "session object."
    
    Thus, the session-related objects handled by this program are:
        class SessionLog
        class SessionRecording
        class SessionDownloadFile
        
    SessionLog
        This is a collection of SessionLogEntry objects. In other words, the session log file is parsed
        into records, very similarly to the way the regular log files are parsed, but the records from
        a given session log are kept together.
        
    SessionRecording
        This is a plain-text file name, and "black box" contents. When the SessionRecording object is initialized
        from a recording file, the contents are base64-encoded and stored, without any provision for allowing
        ElasticSearch queries of the contents.
        
    SessionDownloadFile
        This is handled the same way as the session recording -- the contents of each file are base64-encoded, but
        their names are stored as plain text. Each file is handled separately -- to see which ones were downloaded
        as part of which session, it's necessary to look at the session logs or play back the recordings from
        the relevant dates and times.
        
   
""" 




import abc
import os
import os.path
#from os import walk, remove, stat, listdir #, sep 
#from os.path import join, isdir, os.path.islink, realpath
import re
import datetime
import time

from pogo.file.stretch_file import AttemptFile, LogFile
from pogo.file.stretch_file import  SessionLogFile, SessionDownloadFile, SessionRecordingFile

"""
    Base class for objects that keep track of lists of files of one specific
    HonSSH-generated type, for example attempt files or session recordings.
    Child classes differentiate themselves by implementing the methods decorated
    with @abc.abstractmethod, and optionally overriding other methods, such
    as my_directories().
"""
class FileLister(object):
    __metaclass__ = abc.ABCMeta
    
    DONE_EXTENSION = '.DONE'

    # Some static methods
    @staticmethod
    def done_name(stretch_file_name):
        return stretch_file_name + FileLister.DONE_EXTENSION
    
    @staticmethod
    def is_done(stretch_file_name):
        # return True if there's a file named filespec + '.DONE'
        return os.path.isfile(FileLister.done_name(stretch_file_name)) and not FileLister.is_a_done_marker(stretch_file_name)
    
    """
        Does this file mark another file as "done," or does it
        simply happen to end with DONE_EXTENSION? (That would
        be legal for a session download file.)
    """
    @staticmethod
    def is_a_done_marker(name):
        # Is there another file with the same name as this one, minus DONE_EXTENSION?
        return os.path.isfile(name[0:len(name) - len(FileLister.DONE_EXTENSION)]) and name.endswith(FileLister.DONE_EXTENSION) 
    
    @staticmethod
    def mark_as_done(stretch_file_object):
        # open().close() is equivalent of 'touch' - creates
        # file if it doesn't already exist.
        fname = FileLister.done_name(stretch_file_object.name())
        open(fname, 'a').close()


    
    def __init__(self, source_dir):
        if not os.path.isdir(source_dir):
            raise ValueError(source_dir + ' is not a directory.')
        else:
            self.source_dir = source_dir
            self._pending_file_names = []
            self._done_file_names = []
            self._pending_file_objects = []
            self.set_latest_timestamp_to_process()
            
    """
        Generate a list of directories to check for our files
    """
    def my_directories(self):
        dirlist = [self.source_dir]
        for (root, dirs, files) in os.walk(self.source_dir):
            dirlist += [ os.path.join(root,dir) for dir in dirs ]
        return dirlist
    
    """
        Given a file name, is this a file that this file lister should process?
    """
    def one_of_my_files(self, name):
        return self.get_filespec_pattern().match(name)

    """
        Generate two lists of files: 1) files that still need
        to be processed, and 2) files that are done.
    """
    def load_file_name_lists(self):
        self._pending_file_names = []
        self._done_file_names = []
        for dn in self.my_directories():
            for name in os.listdir(dn):
                if not self.one_of_my_files(name): continue
                whole_path = os.path.join(dn, name)
                if not os.path.isfile(whole_path): continue #isfile() returns True for symlinks to files.
                if os.path.islink(whole_path): # If it is a symlink, use the actual file instead.
                    path_to_add = os.path.realpath(whole_path)
                else:
                    path_to_add = whole_path
                # Don't process the files that are there merely to mark another file as "done."
                if FileLister.is_a_done_marker(path_to_add): continue
                # Don't process files that have already been done - add these to the "done" list instead.
                if self.is_done(path_to_add):
                    self._done_file_names.append(path_to_add)
                else:
                    self._pending_file_names.append(path_to_add)
        # Eliminate duplicates -- possible with symlinks:
        self._pending_file_names = list(set(self._pending_file_names))
        self._pending_file_names.sort() # simple ascending sort of names
        self._done_file_names = list(set(self._done_file_names))
        self._done_file_names.sort()


    @abc.abstractmethod
    def get_file_class(self):
        return ''
    
    @abc.abstractmethod
    def get_filespec_pattern(self):
        return ''
    
    
    def __iter__(self):
        return iter(self._pending_file_objects)
    
    def __len__(self):
        return len(self._pending_file_objects)
        
    """
        Get the timestamp of 23:59:59: last night. We
        don't want to process any files unless they're
        earlier than that.
    """
    def set_latest_timestamp_to_process(self):
        date_yesterday = datetime.date.today() - datetime.timedelta(days=1)
        mn_last_night= datetime.datetime.combine(date_yesterday, datetime.time(23,59,59) )
        self._latest_timestamp_to_process = time.mktime(mn_last_night.timetuple())
        
    def load_pending_file_objects(self):
        # Don't process files from today...
        file_class = self.get_file_class()
        self._pending_file_objects = []
        for name in self._pending_file_names:
            file_mtime = os.stat(name).st_mtime
            if file_mtime < self._latest_timestamp_to_process:
                self._pending_file_objects.append(file_class(name))
            
    def delete_done_files(self):
        count_removed = 0
        for name in self._done_file_names:
            try:
                os.remove(name)
                os.remove(self.done_name(name))
                count_removed += 1
            except OSError, e:  ## if failed, report it back to the user ##
                print "Error: {0} - {1}.".format(e.filename,e.strerror)
        return count_removed

            
class AttemptFileLister(FileLister):
    
    FILESPEC_PATTERN=re.compile('^\d{8}$')
    
    def __init__(self, source_dir):
        super( AttemptFileLister, self ).__init__(source_dir)
    
    def get_file_class(self):
        return AttemptFile
    
    def get_filespec_pattern(self):
        return AttemptFileLister.FILESPEC_PATTERN


class LogFileLister(FileLister):
    FILESPEC_PATTERN = re.compile('^honssh\.log.*')

    def __init__(self, source_dir):
        super( LogFileLister, self ).__init__(source_dir)

    def get_file_class(self):
        return LogFile
    
    def get_filespec_pattern(self):
        return LogFileLister.FILESPEC_PATTERN
   
 

class SessionLogFileLister(FileLister):
    FILESPEC_PATTERN = re.compile('.*\.log') # This regex pattern will stop working in 3000 AD!
    def __init__(self, source_dir):
        super( SessionLogFileLister, self ).__init__(source_dir)
    
    """
        Override my_directories because we don't search recursively for this kind of file
    """
    def my_directories(self):
        dirs = [ os.path.join(self.source_dir, d) for d in os.listdir(self.source_dir) if os.path.isdir(os.path.join(self.source_dir, d)) ]
        return dirs
    
    def get_file_class(self):
        return SessionLogFile
    
    def get_filespec_pattern(self):
        return SessionLogFileLister.FILESPEC_PATTERN
    

class SessionRecordingFileLister(FileLister):
    FILESPEC_PATTERN = re.compile('.*\.tty') 
    def __init__(self, source_dir):
        super( SessionRecordingFileLister, self ).__init__(source_dir)
    
    def get_file_class(self):
        return SessionRecordingFile
    
    def get_filespec_pattern(self):
        return SessionRecordingFileLister.FILESPEC_PATTERN
    
    """
        Override my_directories because we don't search recursively for this kind of file
    """
    def my_directories(self):
        dirs = [ os.path.join(self.source_dir, d) for d in os.listdir(self.source_dir) if os.path.isdir(os.path.join(self.source_dir, d)) ]
        return dirs
   
class SessionDownloadFileLister(FileLister):
    FILESPEC_PATTERN = None # This pattern is not relevant for htis class
    def __init__(self, source_dir):
        super( SessionDownloadFileLister, self ).__init__(source_dir)
    
    def get_file_class(self):
        return SessionDownloadFile
    
    def get_filespec_pattern(self):
        return SessionDownloadFileLister.FILESPEC_PATTERN

    """
        Override my_directories because we don't search recursively for this kind of file
    """
    def my_directories(self):
        dirs = [ os.path.join(self.source_dir, d, 'downloads') for d in os.listdir(self.source_dir) if os.path.isdir(os.path.join(self.source_dir, d, 'downloads')) ]
        return dirs

    """
        Override this for this class -- these file names can be any file name
        legal in the system.
    """
    def one_of_my_files(self, name):
        return True




"""
    Interface with classes used to communicate with local storage (sqlite database)
"""

class ServiceLocal(object):
    def __init__(self, dao_object):
        if dao_object is None:
            raise ValueError("ServiceLocal class needs a dao local object")
        self._do = dao_object
        
    def get_non_processed(self):
        return self._do.list_where("es_id = ''")
    
    def update_with_es_id(self, db_id, es_id):
        where_clause = "db_id = " + str(db_id)
        fields = ( 'es_id', )
        new_values = ( es_id, )
        self._do.update_where(fields, new_values, where_clause)
    
    def delete_finished_records(self):
        return self._do.delete_where("es_id != ''")
    
    def write_new_records(self, records):
        return self._do.insert_bulk(records)
        
    def write_single_record(self, record):
        self._do.insert_single(record)
        
        


'''
# boing: Main module
# 
# Copyright 2015, Tony Rein
# Licensed under MIT.
# '''
# 
# 
# def main():
#     '''
#     Main function of the boilerplate code is the entry point of the 'boing' executable script (defined in setup.py).
#     
#     Use doctests, those are very helpful.
#     
#     >>> main()
#     Hello
#     >>> 2 + 2
#     4
#     '''
#     
#     print("Hello")




'''
boing: Boing is a utility for putting data generated by a HonSSH honeypot into an Elasticsearch database.

Note that "python setup.py test" invokes pytest on the package. With appropriately
configured setup.cfg, this will check both xxx_test modules and docstrings.

Copyright 2015, Tony Rein.
Licensed under MIT.
'''
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from setuptools.command.install import install as _install
from pkg_resources import resource_stream
import shutil



# This is a plug-in for setuptools that will invoke py.test
# when you run python setup.py test
class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest  # import here, because outside the required eggs aren't loaded yet
        sys.exit(pytest.main(self.test_args))


version = "0.9.3.6b"

class install(_install):
    def install_config_file(self):
        instream = resource_stream('pogo', '/data/pogo.cfg')
        outfilename = '/etc/pogo.cfg'
        with open(outfilename, 'wt') as f:
            shutil.copyfileobj(instream, f)
    def install_logrotate_conf(self):
        instream = resource_stream('pogo', '/data/logrotate.cfg')
        outfilename = '/etc/logrotate.d/pogo'
        with open(outfilename, 'wt') as f:
            shutil.copyfileobj(instream, f)

    def run(self):
        _install.run(self)
        print 'Post-install stage...'
        print 'Copying config file...'
        self.install_config_file()
        self.install_logrotate_conf()

setup(name="pogo",
      version=version,
      description="Pogo is a utility for putting data generated by a HonSSH honeypot into an Elasticsearch database.",
      long_description=open("README.rst").read(),
      classifiers=[ # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Topic :: Security',
        'Topic :: Utilities',
        'Intended Audience :: System Administrators'
      ],
      keywords="HonSSH Elasticsearch Python Honeypot cybersecurity", # Separate with spaces
      author="Tony Rein",
      author_email="boing.to.elasticsearch@gmail.com",
      url="https://github.com/tonyrein/pogo.git",
      license="MIT",
      packages=find_packages(exclude=['examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      tests_require=['pytest'],
      cmdclass={'test': PyTest, 'install': install},
      package_data = {'pogo': ['data/pogo_schema.sql', 'data/pogo.cfg', 'data/logrotate.cfg']},
      install_requires=['iso8601', 'tzlocal', 'python-geoip', 'python-geoip-geolite2', 'elasticutils', 'elasticsearch'],
      # TODO: List executable scripts, provided by the package (this is just an example)
      entry_points={
        'console_scripts':
            ['pogo=pogo.main:main']
      }
)



import abc
import sqlite3

import os
import os.path
import sys
from pogo.dao.local_db_access import LocalDBAccessor
from pogo.util.config import StretchConfig

class RecordDaoLocal(object):
    __metaclass__ = abc.ABCMeta
    def __init__(self, localdbaccessor):
        if not localdbaccessor:
            raise ValueError("RecordDaoLocal object needs a LocalDBAccessor.")
        else:
            self._dba = localdbaccessor
            
        """
        assemble a SQL insert string appropriate for
        this object's class.
        The result will be something like
        "INSERT INTO attempts ( timestamp, bifrozt_host, source_ip ) VALUES ( ?,?,? )"
    """    
    def build_insert_query(self):
        table_name = self.get_table_name()
        insert_fields = self.get_insert_fields()
        sql = "INSERT INTO " + table_name + " ( "
        sql +=  ','.join(insert_fields) + " ) VALUES ( "
        sql += len(insert_fields) * '?,'
        if sql.endswith(','):
            sql = sql.rstrip(',')
        sql += " )"
        return sql
    
    """
        Make a list of values for a record, with the fields
        in the same order as in this class's INSERT_FIELDS tuple.
    """
    def build_values_list(self, record):
        d = record.as_dict()
        ret_list = []
        for f in self.get_insert_fields():
            ret_list.append(d[f])
        return ret_list

             
    """
        Insert a single record.
    """
    def insert_single(self, record ):
        sql = self.build_insert_query()
        values_list = self.build_values_list(record)
        try:
#             self.db_open()
            cursor = self._dba.db.cursor()
            cursor.execute('BEGIN TRANSACTION')
            cursor.execute(sql, values_list )
            cursor.execute('COMMIT')
        except sqlite3.Error as e:  # @UndefinedVariable
            cursor.execute('ROLLBACK')
            raise e
        
    """
        records should be a tuple of dto records, all of
        the type appropriate for this record_dao_local object.
        That is, if this is an AttemptRecordDaoLocal, all
        the records should be instances of AttemptRecord
    """
    def insert_bulk(self, records):
        sql = self.build_insert_query()
        count_of_written = 0
        try:
            cursor = self._dba.db.cursor()
            cursor.execute('BEGIN TRANSACTION')
            for r in records:
                values_list = self.build_values_list(r)
                cursor.execute(sql, values_list)
                count_of_written+=1
            cursor.execute('COMMIT')
            return count_of_written
        except sqlite3.Error as e:  # @UndefinedVariable
            cursor.execute('ROLLBACK')
            raise e

    
    def list_all(self):
        return self.list_where(None)
        
    def list_where(self, where_clause=None):
        sql = "SELECT " + self.get_all_fields() + " FROM " + self.get_table_name()
        if where_clause:
            sql += " WHERE " + where_clause
        cursor = self._dba.db.cursor()
        cursor.execute(sql)
        return cursor.fetchall()
    
    """
        Meant to be used as follows:
        new_values = ('a1', 'a2')
        where_clause = "db_id = 2444"
        fields = ('columnA', 'columnII')
        dao_object.update_where(fields, new_values, where_clause)
        Of course, fields and corresponding values must be in
        the same order.
    """
    def update_where(self, fields, new_values, where_clause=None):
        table_name = self.get_table_name()
        sql = "UPDATE " + table_name + " SET "
        if fields and len(fields) > 0:
            for f in fields:
                sql += "'" + f + "' =  ?,"
            # remove last comma, if present
            if sql.endswith(','):
                sql = sql.rstrip(',')
        if where_clause is not None:
            sql += " WHERE " + where_clause
        try:
            cursor = self._dba.db.cursor()
            cursor.execute('BEGIN TRANSACTION')
            cursor.execute(sql, new_values)
            cursor.execute('COMMIT')
        except sqlite3.Error as e:  # @UndefinedVariable
            cursor.execute('ROLLBACK')
            raise e
        
    def delete_where(self, where_clause):
        if not where_clause:
            raise ValueError("RecordDaoLocal.delete_where() called without where clause.")
        table_name = self.get_table_name()
        sql = "DELETE FROM " + table_name + " WHERE " + where_clause
        count_deleted = 0
        try:
            cursor = self._dba.db.cursor()
            cursor.execute('BEGIN TRANSACTION')
            cursor.execute(sql)
            count_deleted = cursor.rowcount
            cursor.execute('COMMIT')
            return count_deleted # should be the number of rows changed
        except sqlite3.Error as e:  # @UndefinedVariable
            cursor.execute('ROLLBACK')
            raise e


             
    @abc.abstractmethod
    def get_table_name(self):
        return ''
    
    @abc.abstractmethod
    def get_all_fields(self):
        return ''
    
    @abc.abstractmethod
    def get_insert_fields(self):
        return ''

class AttemptRecordDaoLocal(RecordDaoLocal):
    TABLE_NAME = 'attempts'
    ALL_FIELDS = "db_id, es_id, timestamp, bifrozt_host, source_ip, user, password, success, country_code, country_name"
    INSERT_FIELDS = ( 'timestamp',  'bifrozt_host',
                       'source_ip', 'user', 'password', 'success',
                       'country_code', 'country_name' )
    def __init__(self, localdbaccessor):
        super(AttemptRecordDaoLocal,self).__init__(localdbaccessor)
        
    def get_table_name(self):
        return AttemptRecordDaoLocal.TABLE_NAME
    
    def get_all_fields(self):
        return AttemptRecordDaoLocal.ALL_FIELDS
    
    def get_insert_fields(self):
        return AttemptRecordDaoLocal.INSERT_FIELDS
    
        

class LogRecordDaoLocal(RecordDaoLocal):
    TABLE_NAME = 'log_msg'
    ALL_FIELDS = "db_id, es_id, timestamp, bifrozt_host, server_info, message"
    INSERT_FIELDS = ( 'timestamp', 'bifrozt_host', 'server_info', 'message' )
    def __init__(self, localdbaccessor):
        super(LogRecordDaoLocal,self).__init__(localdbaccessor)
        
    def get_table_name(self):
        return LogRecordDaoLocal.TABLE_NAME
    
    def get_all_fields(self):
        return LogRecordDaoLocal.ALL_FIELDS
    
    def get_insert_fields(self):
        return LogRecordDaoLocal.INSERT_FIELDS
    
class SessionLogDaoLocal(RecordDaoLocal):
    TABLE_NAME = 'session_log_records'
    ALL_FIELDS = ALL_FIELDS = "db_id, es_id, timestamp, bifrozt_host, source_ip, country_code, country_name, channel, message"
    INSERT_FIELDS = ( 'timestamp',  'bifrozt_host',
                       'source_ip', 'country_code', 'country_name', 'channel', 'message' )
        
    def __init__(self, localdbaccessor):
        super(SessionLogDaoLocal,self).__init__(localdbaccessor)
        
    def get_table_name(self):
        return SessionLogDaoLocal.TABLE_NAME
    
    def get_all_fields(self):
        return SessionLogDaoLocal.ALL_FIELDS
    
    def get_insert_fields(self):
        return SessionLogDaoLocal.INSERT_FIELDS


class SessionRecordingDaoLocal(RecordDaoLocal):
    TABLE_NAME = 'session_recordings'
    ALL_FIELDS = ALL_FIELDS = "db_id, es_id, timestamp, bifrozt_host, source_ip, country_code, country_name, filename, contents"
    INSERT_FIELDS = ( 'timestamp',  'bifrozt_host',
                       'source_ip', 'country_code', 'country_name', 'filename', 'contents' )
        
    def __init__(self, localdbaccessor):
        super(SessionRecordingDaoLocal,self).__init__(localdbaccessor)
        
    def get_table_name(self):
        return SessionRecordingDaoLocal.TABLE_NAME
    
    def get_all_fields(self):
        return SessionRecordingDaoLocal.ALL_FIELDS
    
    def get_insert_fields(self):
        return SessionRecordingDaoLocal.INSERT_FIELDS


class SessionDownloadDaoLocal(RecordDaoLocal):
    TABLE_NAME = 'session_downloads'
    ALL_FIELDS = ALL_FIELDS = "db_id, es_id, timestamp, bifrozt_host, source_ip, country_code, country_name, filename, contents"
    INSERT_FIELDS = ( 'timestamp',  'bifrozt_host',
                       'source_ip', 'country_code', 'country_name', 'filename', 'contents' )
        
    def __init__(self, localdbaccessor):
        super(SessionDownloadDaoLocal,self).__init__(localdbaccessor)
        
    def get_table_name(self):
        return SessionDownloadDaoLocal.TABLE_NAME
    
    def get_all_fields(self):
        return SessionDownloadDaoLocal.ALL_FIELDS
    
    def get_insert_fields(self):
        return SessionDownloadDaoLocal.INSERT_FIELDS



